#include <variable.h>
#include <opengl.h>
#include <vector>
#include <math.h>
#include <iostream>
#include <object.h>

void draw_2p_circle(double xc, double yc, double xe, double ye);
void draw_3p_arc(double xs, double ys, double xc, double yc, double xe, double ye);
void draw_spline(std::vector<std::vector<double>> C);


void display(){

    //int fragments=50; //circle and arc resolution

    //line preview function
    if(object.type=="line"){
        glLineWidth(1);
        glColor4d(1,1,1,1);
        glBegin(GL_LINES);
        glVertex3d(object.start.x,object.start.y,object.start.z);
        glVertex3d(object.end.x,object.end.y,object.end.z);
        glEnd();
    }

    //linestrip preview function
    if(object.type=="linestrip" && object.control.size()>0){
        glLineWidth(1);
        glColor4d(1,1,1,1);
        glBegin(GL_LINE_STRIP);
        glVertex3d(object.start.x,object.start.y,object.start.z);
        for(int i=0; i<object.control.size(); i++){
            glVertex3d(object.control.at(i).x,object.control.at(i).y,object.control.at(i).z);
        }
        glVertex3d(object.end.x,object.end.y,object.end.z);
        glEnd();
    }

    //rectangular preview function
    if(object.type=="rectangular"){
        glLineWidth(1);
        glColor4d(1,1,1,1);
        glBegin(GL_LINE_LOOP);
        glVertex3d(object.start.x,object.start.y,object.start.z); //lower left
        glVertex3d(object.control.front().x,object.control.front().y,object.control.front().z); //upper left
        glVertex3d(object.end.x,object.end.y,object.end.z); //upper right
        glVertex3d(object.control.back().x,object.control.back().y,object.control.back().z); //lower right
        glEnd();
    }

    //circle preview function
    if(object.type=="circle"){
        draw_2p_circle(object.center.x,object.center.y,object.end.x,object.end.y);
    }

    //arc preview function
    if(object.type=="arc" && object.control.size()>0){
        draw_3p_arc(object.start.x, object.start.y, object.control.back().x, object.control.back().y, object.end.x, object.end.y);
    }

    //spline preview function
    if(object.type=="spline" && object.control.size()>0){
        std::vector<std::vector<double>> C; //2d array
        C.push_back({object.start.x, object.start.y}); //spline startpoint
        for(int i=0; i<object.control.size(); i++){
            C.push_back({object.control.at(i).x, object.control.at(i).y}); //spline controlpoints
        }
        C.push_back({object.end.x, object.end.y}); //spline endpoint
        draw_spline(C);
        C.clear();
    }

    //display stored object data
    for(int i=0; i<objects.size(); i++){
        if(objects.at(i).type=="line"){
            glLineWidth(1);
            glColor4d(1,1,1,1);
            glBegin(GL_LINES);
            glVertex3d(objects.at(i).start.x,objects.at(i).start.y,objects.at(i).start.z);
            glVertex3d(objects.at(i).end.x,objects.at(i).end.y,objects.at(i).end.z);
            glEnd();
        } else if(objects.at(i).type=="linestrip"){
            glLineWidth(1);
            glColor4d(1,1,1,1);
            glBegin(GL_LINE_STRIP);
            glVertex3d(objects.at(i).start.x,objects.at(i).start.y,objects.at(i).start.z);
            for(int j=0; j<objects.at(i).control.size(); j++){
                glVertex3d(objects.at(i).control.at(j).x,objects.at(i).control.at(j).y,objects.at(i).control.at(j).z);
            }
            glVertex3d(objects.at(i).end.x,objects.at(i).end.y,objects.at(i).end.z);
            glEnd();
        } else if(objects.at(i).type=="spline"){
            std::vector<std::vector<double>> C; //2d array
            C.push_back({objects.at(i).start.x, objects.at(i).start.y}); //spline startpoint
            for(int j=0; j<objects.at(i).control.size(); j++){
                C.push_back({objects.at(i).control.at(j).x, objects.at(i).control.at(j).y}); //spline controlpoints
            }
            C.push_back({objects.at(i).end.x, objects.at(i).end.y}); //spline endpoint
            draw_spline(C);
            C.clear();
        } else if(objects.at(i).type=="arc"){
            draw_3p_arc(objects.at(i).start.x, objects.at(i).start.y, objects.at(i).control.back().x, objects.at(i).control.back().y, objects.at(i).end.x, objects.at(i).end.y);
        } else if(objects.at(i).type=="circle"){
            draw_2p_circle(objects.at(i).center.x,objects.at(i).center.y,objects.at(i).end.x,objects.at(i).end.y);
        } else if(objects.at(i).type=="rectangular"){
            glLineWidth(1);
            glColor4d(1,1,1,1);
            glBegin(GL_LINE_LOOP);
            glVertex3d(objects.at(i).start.x,objects.at(i).start.y,objects.at(i).start.z); //lower left
            glVertex3d(objects.at(i).control.front().x,objects.at(i).control.front().y,objects.at(i).control.front().z); //upper left
            glVertex3d(objects.at(i).end.x,objects.at(i).end.y,objects.at(i).end.z); //upper right
            glVertex3d(objects.at(i).control.back().x,objects.at(i).control.back().y,objects.at(i).control.back().z); //lower right
            glEnd();
        }
    }
}

void draw_2p_circle(double xc, double yc, double xe, double ye){

        int segments = 25;
        double radius = sqrt(pow(xe-xc,2)+pow(ye-yc,2));
        glColor4d(1,1,1,1);
        glBegin(GL_LINE_LOOP);
        for (int j = 0; j<segments; j++)
        {
            double theta = 2.0 * M_PI * int(j) / int(segments);
            double x = radius * cosf(theta);
            double y = radius * sinf(theta);
            glVertex2d(x + xc, y + yc);
        }
        glEnd();
}

void draw_3p_arc(double xs, double ys, double xc, double yc, double xe, double ye){

    double x1 = (xc + xs) / 2;
    double y1 = (yc + ys) / 2;
    double dy1 = xc - xs;
    double dx1 = -(yc - ys);

    // Get the perpendicular bisector of (x2, y2) and (x3, y3).
    double x2 = (xe + xc) / 2;
    double y2 = (ye + yc) / 2;
    double dy2 = xe - xc;
    double dx2 = -(ye - yc);

    double endpoint_x0 = x1 + dx1;
    double endpoint_y0 = y1 + dy1;
    double endpoint_x1 = x2 + dx2;
    double endpoint_y1 = y2 + dy2;

    //line 1
    double delta_y0 = endpoint_y0 - y1;
    double delta_x0 = x1 - endpoint_x0;
    double c0 = delta_y0 * x1 + delta_x0 * y1;
    //line 2
    double delta_y1 = endpoint_y1 - y2;
    double delta_x1 = x2 - endpoint_x1;
    double c1 = delta_y1 * x2 + delta_x1 * y2;

    double determinant = delta_y0*delta_x1 - delta_y1*delta_x0;
    double start_angle =0;
    double end_angle=0;
    double radius=0;
    double x_center=0;
    double y_center=0;

    if (determinant == 0) // The lines are parallel.
    {
        // lines are parallel
    }
    else  {
        x_center = (delta_x1*c0 - delta_x0*c1)/determinant;
        y_center = (delta_y0*c1 - delta_y1*c0)/determinant;
        double dx = x_center - xs;
        double dy = y_center - ys;
        radius = sqrt(dx * dx + dy * dy);

        //formula if arc is cw or ccw.. this is nice..
        //To determine which side of the line from A=(x1,y1) to B=(x2,y2) a point P=(x,y) falls on you need to compute the value:-
        //d=(x−x1)(y2−y1)−(y−y1)(x2−x1)
        //If d<0 then the point lies on one side of the line, and if d>0 then it lies on the other side. If d=0 then the point lies exactly line.

        double xp = xe;
        double yp = ye;
        double x1 = xs;
        double y1 = ys;
        double x2 = xc;
        double y2 = yc;

        double d = ((xp-x1)*(y2-y1))-((yp-y1)*(x2-x1));
        //qDebug() << "d :" << d;

        if(d > 0){
            start_angle = 180/M_PI*atan2(ye-y_center, xe-x_center);    //start angle G2
            end_angle = 180/M_PI*atan2(ys-y_center, xs-x_center);    //end angle G2
        }
        if(d < 0){
            start_angle = 180/M_PI*atan2(ys-y_center, xs-x_center);    //start angle G3
            end_angle  = 180/M_PI*atan2(ye-y_center, xe-x_center);    //end angle G3
        }
        if(d == 0){
            //draw straight line
        }
    }

    //draw arc..
    double angle_start = (start_angle/360)*(2*M_PI); //start angle
    if(end_angle==360)
    {
        end_angle=0;   //autocad degree problem..
    }
    double angle_end = (end_angle/360)*(2*M_PI);
    if (angle_end<angle_start)   //this to avoid the start angle is bigger then the end angle value
    {
        angle_end+=2*M_PI;
    }

    double arc_angle = angle_end - angle_start;
    double theta = arc_angle / double(segments - 1);//theta is now calculated from the arc angle instead, the - 1 bit comes from the fact that the arc is open
    double tangetial_factor = tanf(theta);
    double radial_factor = cosf(theta);
    double x = radius * cosf(angle_start);//we now start at the start angle
    double y = radius * sinf(angle_start);

    glColor4d(1,1,1,1);
    glBegin(GL_LINE_STRIP);//since the arc is not a closed curve, this is a strip now
    for(int ii = 0; ii < segments; ii++)
    {
        glVertex2d(x + x_center, y + y_center); //x center, y center..

        float tx = -y;
        float ty = x;

        x += tx * tangetial_factor;
        y += ty * tangetial_factor;

        x *= radial_factor;
        y *= radial_factor;
    }
    glEnd();

}

void draw_spline(std::vector<std::vector<double>> C){
    double S[C.size()-1][2][4];
    double w, b[C.size()], d[C.size()], x[C.size()];
    int i, dim;
    int n = C.size()-1; // number of splines

    for(dim=0; dim<2; dim++)
    {
        // define d[]:
        d[0]=3.0*(C[1][dim]-C[0][dim]);
        for(i=1; i<n; i++)
            d[i]=3.0*(C[i+1][dim]-C[i-1][dim]);
        d[n]=3.0*(C[n][dim]-C[n-1][dim]);
        // forward sweep: (simplified tridiagonal solution: all a[i]=1 and all c[i]=1)
        b[0]=2.0;
        for(i=1; i<C.size(); i++)
        {
            w = 1.0/b[i-1];
            b[i] = 4.0 - w;
            d[i] -= (w*d[i-1]);
        }
        // calculate solution vector x[i] = D[i]:
        // (Wikipedia x[] = Wolfram D[])
        x[n]=d[n]/b[n];
        for(i=n-1; i>=0; i--)
            x[i]=(d[i]-x[i+1])/b[i];
        // calculate spline S[i][dim] a, b, c and d:
        for(i=0; i<n; i++)
        {
            S[i][dim][0]=C[i][dim];
            S[i][dim][1]=x[i];
            S[i][dim][2]=3.0*(C[i+1][dim]-C[i][dim]) - 2.0*x[i] - x[i+1];
            S[i][dim][3]=2.0*(C[i][dim]-C[i+1][dim]) + x[i] + x[i+1];
        }
    }

    glLineWidth(1);
    glColor4d(1,1,1,1);
    glBegin(GL_LINE_STRIP);
    for(int p=0; p<C.size()-1; p++)  //spline points
    {
        for(double t=0; t<1; t+=0.1)  //time 0-1 for each spline
        {
            double px = ((S[p][0][3]*t+S[p][0][2])*t+S[p][0][1])*t+S[p][0][0];
            double py = ((S[p][1][3]*t+S[p][1][2])*t+S[p][1][1])*t+S[p][1][0];
            //std::cout<< "x:" << px << " y:" << py << std::endl;
            glVertex2d(px,py);
        }
    }
    glEnd();

    C.clear(); //empty 2d array
}


//show preview data for line

//    glColor4d(line.red,line.green,line.blue,line.alpha);
//    glBegin(GL_LINES);
//    glVertex3d(line.xstart,line.ystart,line.zstart);
//    glVertex3d(line.xend,line.yend,line.zend);
//    glEnd();

//    //show stored data for line
//    double lwl=0;
//    for(int i=0; i<lines.size(); i++){
//        //select line
//        if(mode=="select"){
//            double dist=(zr-zl)/(pxw/6);
//            for(double t=0; t<1; t+=0.01){
//                double xt=((1-t)*lines.at(i).xstart+(t*lines.at(i).xend)); //https://math.stackexchange.com/questions/175896/finding-a-point-along-a-line-a-certain-distance-away-from-another-point
//                double yt=((1-t)*lines.at(i).ystart)+(t*lines.at(i).yend);
//                double zt=((1-t)*lines.at(i).zstart)+(t*lines.at(i).zend);
//                //look is mx,my is within the range..
//                if(mx<xt+dist && mx>xt-dist && my<yt+dist && my>yt-dist && mz<zt+dist && mz>zt-dist){ //startpoint snap
//                    lwl=1;
//                    DRAW_SNAP(mx,my,0,dist);
//                    break;
//                }
//            }
//        }

//        //draw line
//        glLineWidth(lines.at(i).linewidth+lwl);
//        glColor4d(lines.at(i).red,lines.at(i).green,lines.at(i).blue,lines.at(i).alpha);
//        glBegin(GL_LINES);
//        glVertex3d(lines.at(i).xstart,lines.at(i).ystart,lines.at(i).zstart);
//        glVertex3d(lines.at(i).xend,lines.at(i).yend,lines.at(i).zend);
//        glEnd();

//    }

//show stored data for ellipse
//for(int i=0; i<ellipses.size(); i++){
//    double center_x = ellipses.at(i).xcenter;
//    double center_y = ellipses.at(i).ycenter;

//    //center point to endpoint mayor axis..
//    double radius_x = sqrt(pow(ellipses.at(i).xstart-ellipses.at(i).xcenter,2) + pow(ellipses.at(i).ystart-ellipses.at(i).ycenter,2));
//    //ratio minor axis to mayor axis..
//    double radius_y = ellipses.at(i).ratio*radius_x;

//    double rotate_x;
//    double rotate_y;

//    //calculate angle of ellipse
//    double x1=ellipses.at(i).xstart;
//    double y1=ellipses.at(i).ystart;
//    double radian = atan2(y1 - center_y, x1 - center_x);
//    double angle = radian * (180 / M_PI);
//    if (angle < 0.0)
//        angle += 360.0;

//    glColor4d(ellipses.at(i).red,ellipses.at(i).green,ellipses.at(i).blue,ellipses.at(i).alpha);
//    glBegin(GL_LINE_STRIP);

//    for(double j = ellipses.at(i).start_angle; j<ellipses.at(i).end_angle; j+=0.1)
//    {

//        //non rotated ellipse..
//        double non_rotated_x = /*center_x +*/ cos(j)*radius_x;
//        double non_rotated_y = /*center_y +*/ sin(j)*radius_y;
//        //glVertex2d(center_x + cos(j)*radius_x, center_y + sin(j)*radius_y);

//        //rotated ellipse..
//        rotate_x = center_x  + cos(angle * M_PI / 180.0 )* non_rotated_x - sin(angle * M_PI / 180.0 ) * non_rotated_y;
//        rotate_y = center_y  + sin(angle * M_PI / 180.0 )* non_rotated_x + cos(angle * M_PI / 180.0 ) * non_rotated_y;
//        glVertex2d(rotate_x, rotate_y);
//    }
//    glEnd();
//}


////show stored data for spline
////if(splines.size()>=2){ //splines are a minimum set of 3 coordinates
//std::vector<std::vector<double>> C; //2d array
//for(int j=0; j<splines.size(); j++){

//    if(splines.at(j).is_end==0){
//        C.push_back({splines.at(j).x,splines.at(j).y});
//    }

//    if(splines.at(j).is_end==1)//spline end
//    {
//        C.push_back({splines.at(j).x,splines.at(j).y});

//        double S[C.size()-1][2][4];
//        double w, b[C.size()], d[C.size()], x[C.size()];
//        int i, dim;
//        int n = C.size()-1; // number of splines

//        for(dim=0; dim<2; dim++)
//        {
//            // define d[]:
//            d[0]=3.0*(C[1][dim]-C[0][dim]);
//            for(i=1; i<n; i++)
//                d[i]=3.0*(C[i+1][dim]-C[i-1][dim]);
//            d[n]=3.0*(C[n][dim]-C[n-1][dim]);
//            // forward sweep: (simplified tridiagonal solution: all a[i]=1 and all c[i]=1)
//            b[0]=2.0;
//            for(i=1; i<C.size(); i++)
//            {
//                w = 1.0/b[i-1];
//                b[i] = 4.0 - w;
//                d[i] -= (w*d[i-1]);
//            }
//            // calculate solution vector x[i] = D[i]:
//            // (Wikipedia x[] = Wolfram D[])
//            x[n]=d[n]/b[n];
//            for(i=n-1; i>=0; i--)
//                x[i]=(d[i]-x[i+1])/b[i];
//            // calculate spline S[i][dim] a, b, c and d:
//            for(i=0; i<n; i++)
//            {
//                S[i][dim][0]=C[i][dim];
//                S[i][dim][1]=x[i];
//                S[i][dim][2]=3.0*(C[i+1][dim]-C[i][dim]) - 2.0*x[i] - x[i+1];
//                S[i][dim][3]=2.0*(C[i][dim]-C[i+1][dim]) + x[i] + x[i+1];
//            }
//        }

//        //spline selection
//        double lws=0;
//        if(mode=="select"){
//            for(int p=0; p<C.size()-1; p++)  //spline points
//            {
//                for(double t=0; t<1; t+=0.01)  //time 0-1 for each spline
//                {
//                    double px = ((S[p][0][3]*t+S[p][0][2])*t+S[p][0][1])*t+S[p][0][0];
//                    double py = ((S[p][1][3]*t+S[p][1][2])*t+S[p][1][1])*t+S[p][1][0];
//                    //std::cout<< "x:" << px << " y:" << py << std::endl;
//                    double dist=(zr-zl)/(pxw/6);
//                    if(mx<px+dist && mx>px-dist && my<py+dist && my>py-dist){ //startpoint snap
//                        DRAW_SNAP(px,py,0,dist);
//                        lws=1;
//                        break;
//                    }
//                }
//            }
//        }

//        //spline draw
//        glLineWidth(splines.at(j).linewidth+lws);
//        glColor4d(splines.at(j).red,splines.at(j).green,splines.at(j).blue,splines.at(j).alpha);
//        glBegin(GL_LINE_STRIP);
//        for(int p=0; p<C.size()-1; p++)  //spline points
//        {
//            for(double t=0; t<1; t+=0.1)  //time 0-1 for each spline
//            {
//                double px = ((S[p][0][3]*t+S[p][0][2])*t+S[p][0][1])*t+S[p][0][0];
//                double py = ((S[p][1][3]*t+S[p][1][2])*t+S[p][1][1])*t+S[p][1][0];
//                //std::cout<< "x:" << px << " y:" << py << std::endl;
//                glVertex2d(px,py);
//            }
//        }
//        glEnd();

//        /*glPointSize(2);
//            glBegin(GL_POINTS);
//            for(int i=0; i<=C.size()-1; i++)  //spline points
//            {
//                glVertex2d(C[i][0],C[i][1]);

//            }
//            glEnd();*/

//        C.clear(); //empty 2d array

//    }
//}
